--[[
BASIC BITMAP LOADING SYSTEM
- - - - - - - - - - - - - -

	Special thanks to:
	------------------
	>Leegao - I used part of his struct.lua
	as reference
	
	-Flacko
	9/9/2010
]]

local function ReadInt(file,endian)
	local b = {0,0,0,0}
	for i=1, 4 do
		local a = file:read(1)
		if a then
			b[i] = string.byte(a)
		end
	end
	if endian == "l" then --Little endian reading, for BMP format
		return b[1] + b[2]*256 + b[3]*256e2 + b[4]*256e3
	elseif endian == "b" then --big endian reading, for PNG format
		return b[1]*256e3 + b[2]*256e2 + b[3]*256 + b[4]
	else
		error("ReadInt: Wrong endian string")
	end
end

local function ReadShort(file, endian)
	local b = {0,0}
	for i=1, 2 do
		local a = file:read(1)
		if a then
			b[i] = string.byte(a)
		end
	end
	if endian == "l" then --Little endian reading, for BMP format
		return b[1] + b[2]*256
	elseif endian == "b" then --big endian reading, for PNG format
		return  b[1]*256 + b[2]
	else
		error("ReadShort: Wrong endian string")
	end
end

local function ReadPixel(file,bpp)
	bpp = bpp or 24
	if bpp == 24 then
		return {
			b = string.byte(file:read(1)),
			g = string.byte(file:read(1)),
			r = string.byte(file:read(1))
		}
	elseif bpp==8 then
	--[[Only 24 and 8 bit color depth is supported
	8bit colormode only works for grayscale
	Not indexed color. No improvements planned]]
		local byte = string.byte(file:read(1))
		return {r = byte,g = byte,b = byte}
	end
end

local function ReadPadding(file,w)
	local bound = w%4
	for i=1,bound do
		file:read(1)
	end
end

local mt = {}
mt.__index = mt

function mt:__call(x,y)
	return self.map[y][x]
end

function loadBitmap(dir)
	local BMP = assert(io.open(dir,"rb"),"loadBitmap: Image '"..dir.."' NOT FOUND!")
	
	local newbmp = {fsize,w,h,bpp,typ,dir = dir, map={}}
	
	local ReadInt = function(e) return ReadInt(BMP,e) end
	local ReadShort = function(e) return ReadShort(BMP,e) end
	
	local FILETYPE = BMP:read(2)
	
	if FILETYPE == "BM" then
		newbmp.typ = "bmp"
		newbmp.fsize = ReadInt("l")
		BMP:read(12)
		newbmp.w = ReadInt("l")
		newbmp.h = ReadInt("l")
		BMP:read(2)
		newbmp.bpp = ReadShort("l")
		newbmp.compression = ReadInt("l")
		
		if newbmp.compression == 0 then --Uncompressed bitmaps supported only
			BMP:read(20)
			for y=newbmp.h, 1, -1 do --Since bmps start at the bottom left, we're doing an inverted loop
				newbmp.map[y] = {}
				for x=1,newbmp.w do
					newbmp.map[y][x] = ReadPixel(BMP,newbmp.bpp)
				end
				ReadPadding(BMP,newbmp.w)--padding
			end
		end
		
		BMP:close()
		return setmetatable(newbmp,mt)
	else
		FILETYPE = FILETYPE..BMP:read(2)
		
		if FILETYPE == "\137PNG" then
			BMP:read(8)
			if BMP:read(4) == "IHDR" then
				newbmp.typ = "png"
				newbmp.w = ReadInt("b")
				newbmp.h = ReadInt("b")
				BMP:close()
				return newbmp
			end
		end
		
	end
	error("loadBitmap: Unknown file header: "..dir)
end


--[[
USAGE
-----

The loadBitmap(dir) function will load an image file.
Regardless of it's extension, it will determine
the bitmap's format using it's headers.
Only windows bitmaps (.BMP) and portable networks graphics(.PNG)
are supported.

A successful call to the loadBitmap function will return a table
with the following keys:

typ | String, representing the format of the image in lowercase letters. ex: "png"
w | Number, it is the width in pixels of the image. ex: 5
h | Number, it is the height in pixels of the image. ex: 300

Additionaly, the function will load additional data for bmp files:

fsize | Number, size of the file in bytes. ex: 10035
bpp | Number, bits per pixel. ex: 24
compression | Number, compression type 0=No compression 5=png compression
map | Table, this table contains more tables describing each pixel's RGB values.
      
You can also call the table you've just with an x and y parameters to get info about the color of that pixel
ex:
	bitmap(1,1).r = 255
	bitmap(1,1).g = 0
	bitmap(1,1).b = 255
]]